"""Stub out a Python package."""

from collections.abc import Mapping
from importlib.metadata import entry_points, EntryPoint
import sys
from typing import Optional

from .args import Args, parse_args
from .builder.builder import Builder


def main(args: Optional[Args] = None) -> int:
    """Get the args and run the specified builder."""
    if not args:
        args = parse_args()

    group = entry_points().select(group="skelly.builders")  # type: ignore
    builders: Mapping[str, EntryPoint] = {ept.name: ept for ept in group}

    try:
        builder_class = builders[args["builder"]].load()
    except KeyError:
        print(
            f"builder '{args['builder']}' not registered"
            + f": registered builders: {', '.join(sorted(builders))}",
            file=sys.stderr,
        )
        return 1

    builder: Builder = builder_class(args)
    builder.build()
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



"""Utilities for target builders."""

from abc import ABCMeta, abstractmethod
from collections.abc import Iterable, Mapping
import os
import os.path
import re
import shutil
import tempfile
from types import TracebackType
from typing import Callable, Optional, Pattern, Tuple, Type, TypeVar

from ..args import Args, prompt_args


class Template:
    """
    Context manager for a template directory, which can either be a static directory,
    or an archive that the context manager will unpacked to a temp directory and
    delete when complete.
    """

    def __init__(self, template_path: str) -> None:
        self._template_path = template_path
        self._temp_dir: Optional[str] = None

    def __enter__(self) -> str:
        if os.path.isdir(self._template_path):
            return self._template_path

        self._temp_dir = tempfile.mkdtemp()
        try:
            shutil.unpack_archive(self._template_path, self._temp_dir)
        except BaseException:
            shutil.rmtree(self._temp_dir)
            raise
        return self._temp_dir

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        if self._temp_dir:
            shutil.rmtree(self._temp_dir)


class Builder(metaclass=ABCMeta):  # pylint: disable=too-few-public-methods
    """ABC for builder plugins."""

    @abstractmethod
    def __init__(self, args: Args) -> None:
        self._args = args
        self._name = self._args["builder"]
        self._template_vars = dict(self._args["template_vars"])
        self.template_path = self._args["template_path"]

    @abstractmethod
    def build(self) -> None:
        """Build the project."""

    def _prompt_template_vars(self, spec: Iterable[Tuple[str, str]]) -> None:
        """Prompt for missing template variables."""
        self._template_vars.update(
            prompt_args(
                [
                    (prompt, name, None)
                    for prompt, name in spec
                    if name not in self._template_vars
                ]
            )
        )

    def _assert_template_vars(
        self, required_vars: Iterable[str], strict: bool = False
    ) -> None:
        """Assert the required template vars are present."""
        expected = sorted(required_vars)
        actual = sorted(self._template_vars.keys())
        if strict:
            if not all(exp in actual for exp in expected):
                raise AssertionError(
                    "incorrect template variables"
                    + f": template variables = {actual}"
                    + f": expected variables = {expected}"
                )
        else:
            if expected > actual:
                raise AssertionError(
                    "missing required template variables"
                    + f": template variables = {actual}"
                    + f": expected variables = {expected}"
                )

    def _get_template_var_name(self) -> str:
        """Get or derive the name for use in templates."""
        return self._template_vars.get("name", self._args["name"])

    def _get_template_var_issue_url(self) -> str:
        """Get or derive the issue tracking URL for use in templates."""
        issue_url = self._template_vars.get("issue_url", None)
        if issue_url:
            return issue_url
        repo = self._template_vars.get("repo", None)
        if not repo:
            raise ValueError("missing repo, could not derive issue_url")
        if repo.startswith("https://gitlab.com"):
            return f"{repo}/-/issues"
        if repo.startswith("https://github.com"):
            return f"{repo}/issues"
        raise ValueError(f"unknown repo: {repo}: could not derive issue_url")

    def _fill_target(self, fill: Callable[[str], None]) -> None:
        """Fill a template and copy it to the target."""
        if not self.template_path:
            raise RuntimeError(f"builder '{self._name}' template path not set")

        target = self._args["target"]

        # Template can be copied as the target and filled there since the target
        # doesn't exist yet
        if not os.path.exists(target):
            with Template(self.template_path) as template:
                shutil.copytree(template, target)
            fill(target)
            return

        if not os.path.isdir(target):
            raise NotADirectoryError(
                f"builder '{self._name}' target is not a directory: {target}"
            )

        # Fill the template in a staging directory so the fill function won't modify
        # an items already in the target
        with tempfile.TemporaryDirectory() as stage_dir:
            with Template(self.template_path) as template:
                shutil.copytree(template, stage_dir, dirs_exist_ok=True)
            fill(stage_dir)
            shutil.copytree(stage_dir, target, dirs_exist_ok=True)


ValueT = TypeVar("ValueT")


def create_literal_mapping(
    mapping: Mapping[str, ValueT], prefix: str = "{{", suffix: str = "}}"
) -> Mapping[str, ValueT]:
    """
    Return a copy of a mapping where all the keys have a prefix and suffix added
    to them.
    """
    return {f"{prefix}{key}{suffix}": value for key, value in mapping.items()}


def create_re_any(
    terms: Iterable[str], prefix: str = "{{", suffix: str = "}}"
) -> Pattern[str]:
    """
    Return a compiled regular expression that will match any of an iterable of terms.
    """
    j_ts = "|".join(
        sorted(re.escape(term.strip(prefix).strip(suffix)) for term in terms)
    )
    return re.compile(rf"{re.escape(prefix)}\s*({j_ts})\s*{re.escape(suffix)}")


def map_replace(
    src: str,
    substitutions: Mapping[str, str],
    terms_re: Optional[Pattern[str]] = None,
    prefix: str = "{{",
    suffix: str = "}}",
) -> Tuple[str, int]:
    """Use a mapping to replace keys with values in a str and return the new str."""
    if not substitutions:
        return (src, 0)
    if not terms_re:
        terms_re = create_re_any(substitutions.keys(), prefix, suffix)
    return terms_re.subn(
        lambda m_obj: substitutions[f"{prefix}{m_obj.group(1)}{suffix}"], src
    )


def map_rename(
    filepath: str,
    substitutions: Mapping[str, str],
    terms_re: Optional[Pattern[str]] = None,
    prefix: str = "{{",
    suffix: str = "}}",
) -> str:
    """
    Use a mapping to replace keys with values in a file name, rename the file and
    return the new file name.
    """
    basename = os.path.basename(filepath)
    new_basename, count = map_replace(basename, substitutions, terms_re, prefix, suffix)
    if count:
        new_filepath = os.path.join(os.path.dirname(filepath), new_basename)
        os.rename(filepath, new_filepath)
        return new_filepath
    return filepath


def tree_map_rename(
    src: str, substitutions: Mapping[str, str], terms: Pattern[str]
) -> None:
    """Rename files in a diretory tree using substitutions for placeholder in names."""
    for root, dirs, files in os.walk(src, topdown=False):
        for name in [*files, *dirs]:
            map_rename(os.path.join(root, name), substitutions, terms)


def tree_map_replace(
    src: str, substitutions: Mapping[str, str], terms: Pattern[str]
) -> None:
    """
    Fill in the template variables in files in a directory tree using substitutions.
    """
    for root, _, files in os.walk(src):
        for file in files:
            with open(os.path.join(root, file), "r+", encoding="utf-8") as fileptr:
                template_text = fileptr.read()
                rendered_text, count = map_replace(template_text, substitutions, terms)
                if count:
                    fileptr.seek(0)
                    fileptr.write(rendered_text)
                    fileptr.truncate()



"""Create a Python package."""

import os.path
from typing import Optional, Mapping

from skelly.args import Args
from skelly.builder.builder import (
    Builder,
    create_literal_mapping,
    create_re_any,
    tree_map_rename,
    tree_map_replace,
)
from skelly.builder.python.venv import create_venv


class PythonBuilder(Builder):  # pylint: disable=too-few-public-methods
    """Builds a project to create a Python package."""

    def __init__(self, args: Args) -> None:
        super().__init__(args)

        # User provided template
        if self.template_path:
            return

        # Default template, so prompt for/assert missing required template vars
        if not self._args["silent"]:
            self._prompt_template_vars(
                [
                    ("Author", "author"),
                    ("Author email", "email"),
                    ("Description", "description"),
                    ("Repository", "repo"),
                ]
            )
        self._assert_template_vars(["author", "email", "description", "repo"], True)
        self.template_path = os.path.join(os.path.dirname(__file__), "template.tar")

    def _get_builder_opt_env_dir(self) -> Optional[str]:
        """Get env_dir, if not an absolute path, relative to target."""
        env_dir = self._args["builder_opts"].get("env_dir", None)
        return (
            env_dir
            if not env_dir or os.path.isabs(env_dir)
            else os.path.join(self._args["target"], env_dir)
        )

    def _create_full_template_vars(self) -> Mapping[str, str]:
        """Return the full template vars for this builder."""
        name = self._get_template_var_name()
        return {
            **self._template_vars,
            **{
                "name": name,
                "name_h1": "=" * len(name),
                "issue_url": self._get_template_var_issue_url(),
            },
        }

    def _fill(self, target: str) -> None:
        full_template_vars = self._create_full_template_vars()
        literal_template_vars = create_literal_mapping(full_template_vars)
        terms = create_re_any(literal_template_vars.keys())
        tree_map_rename(target, literal_template_vars, terms)
        tree_map_replace(target, literal_template_vars, terms)
        create_venv(
            target,
            self._get_builder_opt_env_dir(),
            self._args["builder_opts"].get("req_file", None),
        )

    def build(self) -> None:
        self._fill_target(self._fill)



"""Parse and validate command line arguments."""

import argparse
from collections.abc import Iterable, Mapping
import os
from typing import Optional, Tuple, TypedDict


class Args(TypedDict):
    """Args type."""

    silent: bool
    name: str
    template_path: Optional[str]
    template_vars: Mapping[str, str]
    builder_opts: Mapping[str, str]
    builder: str
    target: str


def parse_args() -> Args:
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        "skelly", formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument(
        "-b",
        "--builder",
        default="default",
        help="registered builder that will create the project (default: default)",
    )
    parser.add_argument(
        "-o",
        "--builder-opt",
        action="append",
        nargs=2,
        metavar=("OPTION", "VALUE"),
        help="builder option and value; can be specified multiple times, e.g.,\n"
        + "    -o env_dir '.venv310' -o req_file 'requirements.txt'",
    )
    parser.add_argument(
        "-n", "--name", help="project name (default: derived from target)"
    )
    parser.add_argument(
        "-p", "--template-path", help="project template (default: builder template)"
    )
    parser.add_argument(
        "-t",
        "--template-var",
        action="append",
        nargs=2,
        metavar=("VARIABLE", "VALUE"),
        help="template variable and value; can be specified multiple times, e.g.,\n"
        + "    -t author 'Charlotte Jane' -t email 'cjs@example.com'",
    )
    parser.add_argument(
        "-s",
        "--silent",
        action="store_true",
        help="don't prompt for missing template variables",
    )
    parser.add_argument("-v", "--version", action="version", version="skelly 0.0.1")
    parser.add_argument(
        "target",
        nargs="?",
        default=".",
        help="project directory (default: current directory)",
    )

    args = parser.parse_args()
    target = os.path.abspath(args.target)
    name = args.name if args.name else os.path.basename(target)
    if args.template_var:
        template_vars = {pair[0]: pair[1] for pair in args.template_var}
    else:
        template_vars = {}
    if args.builder_opt:
        builder_opts = {pair[0]: pair[1] for pair in args.builder_opt}
    else:
        builder_opts = {}

    return {
        "silent": args.silent,
        "name": name,
        "template_path": args.template_path,
        "template_vars": template_vars,
        "builder_opts": builder_opts,
        "builder": args.builder,
        "target": target,
    }


PromptSpec = Iterable[Tuple[Optional[str], str, Optional[str]]]


def prompt_args(spec: PromptSpec) -> Mapping[str, str]:
    """Prompt for args with defaults."""
    args = {
        name: input(f"{prompt if prompt else name}: ").strip() or default
        for prompt, name, default in spec
    }
    return {name: value for name, value in args.items() if value}



"""Python venv module library for skelly."""

import os.path
import subprocess
from types import SimpleNamespace
from typing import Optional
import venv


class RequirementsEnvBuilder(venv.EnvBuilder):
    """
    Custom EnvBuilder that pip installs the packages in a requirements file after
    the venv is created.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        req_file: str,
        install_dot: bool = True,
        system_site_packages: bool = False,
        clear: bool = False,
        symlinks: bool = False,
        upgrade: bool = False,
        prompt: Optional[str] = None,
        upgrade_deps: bool = False,
    ) -> None:
        self.req_file = req_file
        self.install_dot = install_dot
        super().__init__(
            system_site_packages,
            clear,
            symlinks,
            upgrade,
            True,  # with_pip is True because we need pip
            prompt,
            upgrade_deps,
        )

    def post_setup(self, context: SimpleNamespace) -> None:
        """
        Run pip to install the packages in the requirments file after the venv
        is created.
        """
        subprocess.run(
            [context.env_exe, "-m", "pip", "install", "-r", self.req_file], check=True
        )
        if self.install_dot:
            subprocess.run(
                [
                    context.env_exe,
                    "-m",
                    "pip",
                    "install",
                    "-e",
                    os.path.dirname(context.env_dir),
                ],
                check=True,
            )


def create_venv(
    target: str,
    env_dir: Optional[str] = None,
    req_file: Optional[str] = None,
    install_dot: bool = True,
) -> None:
    """Create a venv in the target, then pip install the requirements."""
    if not env_dir:
        env_dir = os.path.join(target, ".venv")
    if not req_file:
        req_file = os.path.join(target, "requirements.txt")
    envbuilder = RequirementsEnvBuilder(req_file, install_dot)
    envbuilder.create(env_dir)
